/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

\*---------------------------------------------------------------------------*/

#include "heatFluxFixedValueFvPatchField.H"
#include "volFields.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

template<class Type>
heatFluxFixedValueFvPatchField<Type>::heatFluxFixedValueFvPatchField
(
    const fvPatch& p,
    const Field<Type>& iF
)
:
    fixedValueFvPatchField<Type>(p, iF),
    Tin_(p.size()),
    Tmid_(p.size()),
    Tout_(p.size()),
    k_(0.05),
    rho_(0.021),
    c_(1700),
    dx_(0.001),
    curTimeIndex_(-1)
{}


template<class Type>
heatFluxFixedValueFvPatchField<Type>::heatFluxFixedValueFvPatchField
(
    const heatFluxFixedValueFvPatchField<Type>& ptf,
    const fvPatch& p,
    const Field<Type>& iF,
    const fvPatchFieldMapper& mapper
)
:
    fixedValueFvPatchField<Type>(p, iF),
    Tin_(ptf.Tin_, mapper),
    Tmid_(ptf.Tmid_, mapper),
    Tout_(ptf.Tout_, mapper),
    k_(ptf.k_),
    rho_(ptf.rho_),
    c_(ptf.c_),
    dx_(ptf.dx_),
    curTimeIndex_(-1)
{}


template<class Type>
heatFluxFixedValueFvPatchField<Type>::heatFluxFixedValueFvPatchField
(
    const fvPatch& p,
    const Field<Type>& iF,
    const dictionary& dict
)
:
    fixedValueFvPatchField<Type>(p, iF),
    Tin_("Tin", dict, p.size()),
    Tmid_("Tmid", dict, p.size()),
    Tout_("Tout", dict, p.size()),
    k_(readScalar(dict.lookup("k_solid"))),
    rho_(readScalar(dict.lookup("rho_solid"))),
    c_(readScalar(dict.lookup("c_solid"))),
    dx_(readScalar(dict.lookup("dx_solid"))),
    curTimeIndex_(-1)
{
   if (dict.found("value"))
   {
       fixedValueFvPatchField<Type>::operator==
       (
           Field<Type>("value", dict, p.size())
       );
   }
   else
   {
		scalar alfa2_ = k_/rho_/c_;
		const fvMesh& mesh = p.boundaryMesh().mesh();
		scalar dt_ = mesh.time().deltaT().value();
		//or scalar dt_ = this->db().time.deltaT().value();
		scalar cfl_ = alfa2_*dt_/(dx_*dx_);
		
		heatEq1D(cfl_);
		
		fixedValueFvPatchField<Type>::operator==
		(
			Tout_
		);
	}
}


template<class Type>
heatFluxFixedValueFvPatchField<Type>::heatFluxFixedValueFvPatchField
(
    const heatFluxFixedValueFvPatchField<Type>& ptf,
    const Field<Type>& iF
)
:
    fixedValueFvPatchField<Type>(ptf, iF),
    Tin_(ptf.Tin_),
    Tmid_(ptf.Tmid_),
    Tout_(ptf.Tout_),
    k_(ptf.k_),
    rho_(ptf.rho_),
    c_(ptf.c_),
    dx_(ptf.dx_),
    curTimeIndex_(-1)
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

// Map from self
template<class Type>
void heatFluxFixedValueFvPatchField<Type>::autoMap
(
    const fvPatchFieldMapper& m
)
{
    Field<Type>::autoMap(m);
    Tin_.autoMap(m);
    Tmid_.autoMap(m);
    Tout_.autoMap(m);
}


// Reverse-map the given fvPatchField onto this fvPatchField
template<class Type>
void heatFluxFixedValueFvPatchField<Type>::rmap
(
    const fvPatchField<Type>& ptf,
    const labelList& addr
)
{
    fixedValueFvPatchField<Type>::rmap(ptf, addr);

    const heatFluxFixedValueFvPatchField<Type>& tiptf =
        refCast<const heatFluxFixedValueFvPatchField<Type> >(ptf);

    Tin_.rmap(tiptf.Tin_, addr);
    Tmid_.rmap(tiptf.Tmid_, addr);
    Tout_.rmap(tiptf.Tout_, addr);
}


// Update the coefficients associated with the patch field
template<class Type>
void heatFluxFixedValueFvPatchField<Type>::updateCoeffs()
{
    if (this->updated())
    {
        return;
    }

    if (curTimeIndex_ != this->db().time().timeIndex())
    {
	 	 //get the value of the temperature from inside the domain onto
		 //the boundary
       this->checkVolField();
       fvPatchField<Type>::operator==(this->patchInternalField());
       
	    Field<Type>& patchField = *this;
		 	 
		 scalar alfa2_ = k_/rho_/c_;		 
		 scalar dt_ = this->db().time().deltaT().value();
		 Tout_ = patchField;
		 scalar cfl_ = alfa2_*dt_/(dx_*dx_);
		 
		 heatEq1D(cfl_);
		 
		 patchField = Tout_;
		 
       curTimeIndex_ = this->db().time().timeIndex();
	 }
    
    fixedValueFvPatchField<Type>::updateCoeffs();
}


// Write
template<class Type>
void heatFluxFixedValueFvPatchField<Type>::write(Ostream& os) const
{
    fvPatchField<Type>::write(os);
    Tin_.writeEntry("Tin", os);
    Tmid_.writeEntry("Tmid", os);
    Tout_.writeEntry("Tout", os);
    os.writeKeyword("k_solid")
	      << k_ << token::END_STATEMENT << nl;
    os.writeKeyword("rho_solid")
	      << rho_ << token::END_STATEMENT << nl;
    os.writeKeyword("c_solid")
	      << c_ << token::END_STATEMENT << nl;
    os.writeKeyword("dx_solid")
	      << dx_ << token::END_STATEMENT << nl;
 
    this->writeEntry("value", os);
}

//Integrate 1D heat equation
template<class Type>
void heatFluxFixedValueFvPatchField<Type>::heatEq1D(const scalar& cfl)
{
	if(CFL < cfl)//too high cfl number
	{
		if(4.0 < cfl/CFL)//if it is realy too high then stop
		{
			FatalErrorIn
			(
				"CFL too large in the heat equation:"
			)  << cfl
				<< abort(FatalError);
		}
		else
		{//split the time integration in 4 smaller steps
			scalar cfl_ = 0.25*cfl;
			for(label i = 0; i < 4; i++)
			{
				//1d heat equation
				Tmid_ = Tmid_+cfl_*(Tout_-2*Tmid_+Tin_);
				Tout_ = Tmid_;
			}
		}
	}
	else
	{			
		//1d heat equation
		Tmid_ = Tmid_+cfl*(Tout_-2*Tmid_+Tin_);
		Tout_ = Tmid_;
	}
	return;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
